from dbparti.backends import BasePartition, transaction
from dbparti.backends.utilities import DateTimeUtil
from dbparti.backends.exceptions import PartitionRangeSubtypeError, PartitionFunctionError


"""
MySQL partition backend.

MySQL supports partitioning natively via PARTITION BY clause. Thing to keep in
mind is that python MySQLdb doesn't support compound MySQL statements, so each
separate statement should be executed inside each self.cursor.execute().
"""
class Partition(BasePartition):
    """Common methods for all partition types"""
    def prepare(self):
        """Converts original table to partitioned one"""
        self.cursor.execute("""
            -- We need to rebuild primary key for our partitioning to work
            ALTER table {parent_table} DROP PRIMARY KEY, add PRIMARY KEY ({pk}, {partition_column});
        """.format(
            parent_table=self.table,
            partition_column=self.partition_column,
        ))

        transaction.commit_unless_managed()

    def exists(self):
        """Checks if partition exists"""
        self.cursor.execute("""
            SELECT EXISTS(
                SELECT 1 FROM information_schema.partitions
                WHERE table_name='{parent_table}' AND partition_name='{partition_name}');
        """.format(
            pk = self.partition_pk.column,
            parent_table=self.table,
            partition_name=self._get_name(),
        ))

        return self.cursor.fetchone()[0]


class RangePartition(Partition):
    """Range partition type implementation"""
    def __init__(self, *args, **kwargs):
        super(RangePartition, self).__init__(*args, **kwargs)
        self.partition_range = kwargs['partition_range']
        self.partition_subtype = kwargs['partition_subtype']
        self.datetime = DateTimeUtil(self.column_value, self.partition_range, model=self.model)

    def prepare(self):
        """Converts original table to partitioned one"""
        super(RangePartition, self).prepare()
        self.datetime.now = None

        self.cursor.execute("""
            -- We need to create zero partition to speed up things due to the partitioning
            -- implementation in the early versions of MySQL database (see bug #49754)
            ALTER TABLE {parent_table} PARTITION BY RANGE ({function}({partition_column}))(
                PARTITION {partition_pattern} VALUES LESS THAN (0)
            );
        """.format(
            parent_table=self.table,
            partition_column=self.partition_column,
            partition_pattern=self._get_name(),
            function=self._get_partition_function(),
        ))

        transaction.commit_unless_managed()

    def create(self):
        """Creates new partition"""
        self.cursor.execute("""
            ALTER TABLE {parent_table} ADD PARTITION (
                PARTITION {child_table} VALUES LESS THAN ({function}('{period_end}') + {addition})
            );
        """.format(
            child_table=self._get_name(),
            parent_table=self.table,
            function=self._get_partition_function(),
            period_end=self.datetime.get_period()[1],
            addition='86400' if self._get_column_type() == 'timestamp' else '1',
        ))

        transaction.commit_unless_managed()

    def _get_name(self):
        """Dynamically defines new partition name depending on the partition subtype"""
        try:
            return getattr(self, '_get_{0}_name'.format(self.partition_subtype))()
        except AttributeError:
            import re
            raise PartitionRangeSubtypeError(
                model=self.model,
                current_value=self.partition_subtype,
                allowed_values=[re.match('_get_(\w+)_name', c).group(1) for c in dir(
                    self) if re.match('_get_\w+_name', c) is not None]
            )

    def _get_date_name(self):
        """Defines name for a new partition for date partition subtype"""
        return '{0}_{1}'.format(self.table, self.datetime.get_name())

    def _get_partition_function(self):
        """Returns correct partition function depending on the MySQL column type"""
        functions = {
            'date': 'TO_DAYS',
            'datetime': 'TO_DAYS',
            'timestamp': 'UNIX_TIMESTAMP',
        }

        column_type = self._get_column_type()

        try:
            return functions[column_type]
        except KeyError:
            raise PartitionFunctionError(current_value=column_type, allowed_values=functions.keys())

    def _get_column_type(self):
        """
        We can't rely on self.column_type in MySQL, because Django uses only date
        and datetime types internally, but MySQL has an additional timestamp type
        and we need to know that, otherwise we can apply incorrect partition function
        """
        self.cursor.execute("""
            SELECT data_type
            FROM information_schema.columns
            WHERE table_name = '{parent_table}' AND column_name = '{partition_column}';
        """.format(
            parent_table=self.table,
            partition_column=self.partition_column,
        ))

        return self.cursor.fetchone()[0]



from dbparti import connection, transaction


class BasePartition(object):
    """Base partition class for all backends. All backends should inherit from it."""
    def __init__(self, column_value, column_type, **kwargs):
        self.cursor = connection.cursor()
        self.model = kwargs['object_name']
        self.table = kwargs['db_table']
        self.partition_pk = kwargs['pk']
        self.partition_column = kwargs['partition_column']
        self.column_value = column_value
        self.column_type = column_type

    def prepare(self):
        """Prepares everything that is needed to initialize partitioning"""
        raise NotImplementedError('Prepare method not implemented for partition type: {0}'.format(self.__class__.__name__))

    def exists(self):
        """Checks if partition exists"""
        raise NotImplementedError('Exists method not implemented for partition type: {0}'.format(self.__class__.__name__))

    def create(self):
        """Creates new partition"""
        raise NotImplementedError('Create method not implemented for partition type: {0}'.format(self.__class__.__name__))

    def _get_name(self):
        """Defines name for a new partition"""
        raise NotImplementedError('Name method not implemented for partition type: {0}'.format(self.__class__.__name__))

    def _get_partition_function(self):
        """Contains a partition function that is used to create new partitions at database level"""
        raise NotImplementedError('Partition function method not implemented for partition type: {0}'.format(self.__class__.__name__))


class BasePartitionFilter(object):
    """Base class for all filter types. All filter types should inherit from it."""
    def __init__(self, partition_show, **kwargs):
        self.partition_show = partition_show
        self.model = kwargs['object_name']
        self.table = kwargs['db_table']
        self.partition_column = kwargs['partition_column']

    def apply(self):
        """Contains a filter that needs to be applied to queryset"""
        raise NotImplementedError('Filter not implemented for type: {0}'.format(self.__class__.__name__))



from datetime import datetime, timedelta
from dbparti.backends.exceptions import PartitionRangeError


"""Provides date and time calculations for some database backends"""
class DateTimeUtil(object):
    def __init__(self, now, period, format='%Y-%m-%d %H:%M:%S', model=None):
        self.now = now
        self.period = period
        self.format = format
        self.model = model

    def get_name(self):
        """Returns name of the partition depending on the given date and period"""
        patterns = {
            'day': {'real': 'y%Yd%j', 'none': 'y0000d000'},
            'week': {'real': 'y%Yw%V', 'none': 'y0000w00'},
            'month': {'real': 'y%Ym%m', 'none': 'y0000m00'},
            'year': {'real': 'y%Y', 'none': 'y0000'},
        }

        try:
            return patterns[self.period]['none'] if self.now is None else self.now.strftime(patterns[self.period]['real'])
        except KeyError:
            raise PartitionRangeError(model=self.model, current_value=self.period, allowed_values=patterns.keys())

    def get_period(self):
        """Dynamically returns beginning and an end depending on the given period"""
        try:
            return getattr(self, '_get_{0}_period'.format(self.period))()
        except AttributeError:
            import re
            raise PartitionRangeError(
                model=self.model,
                current_value=self.period,
                allowed_values=[re.match('_get_(\w+)_period', c).group(1) for c in dir(
                    self) if re.match('_get_\w+_period', c) is not None]
            )

    def _get_day_period(self):
        """Returns beginning and an end for a day period"""
        start = self.now.replace(hour=0, minute=0, second=0, microsecond=0)
        end = self.now.replace(hour=23, minute=59, second=59, microsecond=999999)

        return start.strftime(self.format), end.strftime(self.format)

    def _get_week_period(self):
        """Returns beginning and an end for a week period"""
        date_ = datetime(self.now.year, 1, 1)

        if date_.weekday() > 3:
            date_ = date_ + timedelta(7 - date_.weekday())
        else:
            date_ = date_ - timedelta(date_.weekday())

        days = timedelta(days=(int(self.now.strftime('%V')) - 1) * 7)

        start = (date_ + days)
        end = (date_ + days + timedelta(days=6)).replace(hour=23, minute=59, second=59, microsecond=999999)

        return start.strftime(self.format), end.strftime(self.format)

    def _get_month_period(self):
        """Returns beginning and an end for a month period"""
        fday = datetime(self.now.year, self.now.month, 1)

        if self.now.month == 12:
            lday = datetime(self.now.year, self.now.month, 31, 23, 59, 59, 999999)
        else:
            lday = (datetime(self.now.year, self.now.month + 1, 1, 23, 59, 59, 999999) - timedelta(days=1))

        return fday.strftime(self.format), lday.strftime(self.format)

    def _get_year_period(self):
        """Returns beginning and an end for a year period"""
        fday = datetime(self.now.year, 1, 1)
        lday = (datetime(self.now.year + 1, 1, 1, 23, 59, 59, 999999) - timedelta(days=1))

        return fday.strftime(self.format), lday.strftime(self.format)



from django.db import connection, transaction
from dbparti.backends.exceptions import BackendError


try:
    backend = __import__('dbparti.backends.{0}'.format(connection.vendor), fromlist='*')
except ImportError:
    import pkgutil, os
    raise BackendError(
        allowed_values=[name for _, name, is_package in pkgutil.iter_modules(
            [os.path.join(os.path.dirname(__file__), 'backends')]) if is_package]
    )



from django.contrib import admin
from dbparti import backend
from dbparti.backends.exceptions import PartitionColumnError, PartitionFilterError


class PartitionableAdmin(admin.ModelAdmin):
    partition_show = 'all'

    def __init__(self, *args, **kwargs):
        super(PartitionableAdmin, self).__init__(*args, **kwargs)

        if not self.opts.partition_column in self.opts.get_all_field_names():
            raise PartitionColumnError(
                model=self.opts.__dict__['object_name'],
                current_value=self.opts.partition_column,
                allowed_values=self.opts.get_all_field_names()
            )

        try:
            self.filter = getattr(backend.filters, '{0}PartitionFilter'.format(
                self.opts.partition_type.capitalize()))(self.partition_show, **self.opts.__dict__)
        except AttributeError:
            import re
            raise PartitionFilterError(
                model=self.opts.__dict__['object_name'],
                current_value=self.opts.partition_type,
                allowed_values=[c.replace('PartitionFilter', '').lower() for c in dir(
                    backend.filters) if re.match('\w+PartitionFilter', c) is not None and 'Base' not in c]
            )

    def queryset(self, request):
        """Determines data from what partitions should be shown in django admin"""
        qs = super(PartitionableAdmin, self).queryset(request)

        if self.partition_show != 'all':
            qs = qs.extra(where=self.filter.apply())

        return qs



from dbparti import connection


class BasePartitionError(Exception):
    """Base exception class for backend exceptions"""
    def __init__(self, message, **kwargs):
        self.message = message
        self.model = kwargs.get('model', None)
        self.current_value = kwargs.get('current_value', None)
        self.allowed_values = kwargs.get('allowed_values', None)

    def __str__(self):
        return self.message.format(
            model=self.model,
            current=self.current_value,
            vendor=connection.vendor,
            allowed=', '.join(list(self.allowed_values))
        )


class BackendError(BasePartitionError):
    """Unsupported database backend"""
    def __init__(self, **kwargs):
        super(BackendError, self).__init__(
            'Unsupported database backend "{vendor}", supported backends are: {allowed}',
            **kwargs
        )


class PartitionColumnError(BasePartitionError):
    """Undefined partition column"""
    def __init__(self, **kwargs):
        super(PartitionColumnError, self).__init__(
            'Undefined partition column "{current}" in "{model}" model, available columns are: {allowed}',
            **kwargs
        )


class PartitionTypeError(BasePartitionError):
    """Unsupported partition type"""
    def __init__(self, **kwargs):
        super(PartitionTypeError, self).__init__(
            'Unsupported partition type "{current}" in "{model}" model, supported types for "{vendor}" backend are: {allowed}',
            **kwargs
        )


class PartitionFilterError(BasePartitionError):
    """Unsupported partition filter"""
    def __init__(self, **kwargs):
        super(PartitionFilterError, self).__init__(
            'Unsupported partition filter "{current}" in "{model}" model, supported filters for "{vendor}" backend are: {allowed}',
            **kwargs
        )


class PartitionRangeError(BasePartitionError):
    """Unsupported partition range"""
    def __init__(self, **kwargs):
        super(PartitionRangeError, self).__init__(
            'Unsupported partition range "{current}" in "{model}" model, supported partition ranges for backend "{vendor}" are: {allowed}',
            **kwargs
        )


class PartitionRangeSubtypeError(BasePartitionError):
    """Unsupported partition range subtype"""
    def __init__(self, **kwargs):
        super(PartitionRangeSubtypeError, self).__init__(
            'Unsupported partition range subtype "{current}" in "{model}" model, supported range subtypes for backend "{vendor}" are: {allowed}',
            **kwargs
        )


class PartitionShowError(BasePartitionError):
    """Unsupported partition show type"""
    def __init__(self, **kwargs):
        super(PartitionShowError, self).__init__(
            'Unsupported partition show type "{current}" in "{model}" admin class, supported partition show types for backend "{vendor}" are: {allowed}',
            **kwargs
        )


class PartitionFunctionError(BasePartitionError):
    """Unsupported partition function"""
    def __init__(self, **kwargs):
        super(PartitionFunctionError, self).__init__(
            'Unsupported partition function for column type "{current}", supported column types for "{vendor}" backend are: {allowed}',
            **kwargs
        )



from django.db import models
from dbparti import backend
from dbparti.backends.exceptions import PartitionColumnError, PartitionTypeError


models.options.DEFAULT_NAMES += (
    'partition_range',
    'partition_column',
    'partition_type',
    'partition_subtype',
    'partition_list',
)


class Partitionable(models.Model):
    def get_partition(self):
        try:
            field = self._meta.get_field(self._meta.partition_column)
            column_value = field.pre_save(self, self.pk is None)
            column_type = field.get_internal_type()
        except AttributeError:
            raise PartitionColumnError(
                model=self.__class__.__name__,
                current_value=self._meta.partition_column,
                allowed_values=self._meta.get_all_field_names()
            )

        try:
            return getattr(backend.partition, '{0}Partition'.format(
                self._meta.partition_type.capitalize()))(column_value, column_type, **self._meta.__dict__)
        except AttributeError:
            import re
            raise PartitionTypeError(
                model=self.__class__.__name__,
                current_value=self._meta.partition_type,
                allowed_values=[c.replace('Partition', '').lower() for c in dir(
                    backend.partition) if re.match('\w+Partition', c) is not None and 'Base' not in c]
            )

    def save(self, *args, **kwargs):
        partition = self.get_partition()

        if not partition.exists():
            partition.create()

        super(Partitionable, self).save(*args, **kwargs)

    class Meta:
        abstract = True
        partition_type = 'None'
        partition_subtype = 'None'
        partition_range = 'None'
        partition_column = 'None'



from dbparti.models import Partitionable
from django.db.models import get_models
from django.core.management.base import AppCommand


class Command(AppCommand):
    help = 'Configures the database for partitioned models'

    def handle_app(self, app, **options):
        """Configures all needed database stuff depending on the backend used"""
        names = []

        for model in get_models(app):
            if issubclass(model, Partitionable):
                names.append(model.__name__)

                model_instance = model()
                model_instance.get_partition().prepare()

        if not names:
            self.stderr.write('Unable to find any partitionable models in an app: ' + app.__name__.split('.')[0] + '\n')
        else:
            self.stdout.write(
                'Successfully (re)configured the database for the following models: ' + ', '.join(names) + '\n'
            )



from dbparti.backends import BasePartition, transaction
from dbparti.backends.exceptions import (
    PartitionRangeError,
    PartitionRangeSubtypeError
)


"""
PostgreSQL partition backend.

PostgreSQL supports partitioning via inheritance.
"""
class Partition(BasePartition):
    """Common methods for all partition types"""
    def prepare(self):
        """Prepares needed triggers and functions for those triggers"""
        self.cursor.execute("""
            -- We need to create a before insert function
            CREATE OR REPLACE FUNCTION {parent_table}_insert_child()
            RETURNS TRIGGER AS $$
                {partition_function}
            $$ LANGUAGE plpgsql;

            -- Then we create a trigger which calls the before insert function
            DO $$
            BEGIN
            IF NOT EXISTS(
                SELECT 1
                FROM information_schema.triggers
                WHERE event_object_table = '{parent_table}'
                AND trigger_name = 'before_insert_{parent_table}_trigger'
            ) THEN
                CREATE TRIGGER before_insert_{parent_table}_trigger
                    BEFORE INSERT ON {parent_table}
                    FOR EACH ROW EXECUTE PROCEDURE {parent_table}_insert_child();
            END IF;
            END $$;

            -- Then we create a function to delete duplicate row from the master table after insert
            CREATE OR REPLACE FUNCTION {parent_table}_delete_master()
            RETURNS TRIGGER AS $$
                BEGIN
                    DELETE FROM ONLY {parent_table} WHERE {pk} = NEW.{pk};
                    RETURN NEW;
                END;
            $$ LANGUAGE plpgsql;

            -- Lastly we create the after insert trigger that calls the after insert function
            DO $$
            BEGIN
            IF NOT EXISTS(
                SELECT 1
                FROM information_schema.triggers
                WHERE event_object_table = '{parent_table}'
                AND trigger_name = 'after_insert_{parent_table}_trigger'
            ) THEN
                CREATE TRIGGER after_insert_{parent_table}_trigger
                    AFTER INSERT ON {parent_table}
                    FOR EACH ROW EXECUTE PROCEDURE {parent_table}_delete_master();
            END IF;
            END $$;
        """.format(
            pk = self.partition_pk.column,
            parent_table=self.table,
            partition_function=self._get_partition_function()
        ))

        transaction.commit_unless_managed()

    def exists(self):
        """Checks if partition exists. Not used in this backend because everything is done at the database level"""
        return True

    def create(self):
        """Creates new partition. Not used in this backend because everything is done at the database level"""
        pass


class RangePartition(Partition):
    """Range partition type implementation"""
    def __init__(self, *args, **kwargs):
        super(RangePartition, self).__init__(*args, **kwargs)
        self.partition_range = kwargs['partition_range']
        self.partition_subtype = kwargs['partition_subtype']

    def _get_partition_function(self):
        """Dynamically loads needed before insert function body depending on the partition subtype"""
        try:
            return getattr(self, '_get_{0}_partition_function'.format(self.partition_subtype))()
        except AttributeError:
            import re
            raise PartitionRangeSubtypeError(
                model=self.model,
                current_value=self.partition_subtype,
                allowed_values=[re.match('_get_(\w+)_partition_function', c).group(1) for c in dir(
                    self) if re.match('_get_\w+_partition_function', c) is not None]
            )

    def _get_date_partition_function(self):
        """Contains a before insert function body for date partition subtype"""
        patterns = {
            'day': '"y"YYYY"d"DDD',
            'week': '"y"IYYY"w"IW',
            'month': '"y"YYYY"m"MM',
            'year': '"y"YYYY',
        }

        try:
            partition_pattern = patterns[self.partition_range]
        except KeyError:
            raise PartitionRangeError(model=self.model, current_value=self.partition_range, allowed_values=patterns.keys())

        return """
            DECLARE tablename TEXT;
            DECLARE columntype TEXT;
            DECLARE startdate TIMESTAMP;
            BEGIN
                startdate := date_trunc('{partition_range}', NEW.{partition_column});
                tablename := '{parent_table}_' || to_char(NEW.{partition_column}, '{partition_pattern}');

                IF NOT EXISTS(
                    SELECT 1 FROM information_schema.tables WHERE table_name=tablename)
                THEN
                    SELECT data_type INTO columntype
                    FROM information_schema.columns
                    WHERE table_name = '{parent_table}' AND column_name = '{partition_column}';

                    EXECUTE 'CREATE TABLE ' || tablename || ' (
                        CHECK (
                            {partition_column} >= ''' || startdate || '''::' || columntype || ' AND
                            {partition_column} < ''' || (startdate + '1 {partition_range}'::interval) || '''::' || columntype || '
                        )
                    ) INHERITS ({parent_table});';

                    EXECUTE 'CREATE INDEX ' || tablename || '_{partition_column} ON ' || tablename || ' ({partition_column});';
                END IF;

                EXECUTE 'INSERT INTO ' || tablename || ' VALUES (($1).*);' USING NEW;
                RETURN NEW;
            END;
        """.format(
            parent_table=self.table,
            partition_range=self.partition_range,
            partition_column=self.partition_column,
            partition_pattern=partition_pattern
        )



from setuptools import setup, find_packages


setup(
    name='django-db-parti',
    version='0.3.2',
    packages=find_packages(),
    url='https://github.com/maxtepkeev/django-db-parti',
    license=open('LICENSE').read(),
    author='Max Tepkeev',
    author_email='tepkeev@gmail.com',
    description='Fully automatic database table partitioning for Django',
    long_description=open('README.rst').read() + '\n\n' +
                     open('CHANGELOG.rst').read(),
    keywords='django,partition,database,table',
    install_requires=['Django >= 1.4'],
    zip_safe=False,
    classifiers=[
        'Framework :: Django',
        'Development Status :: 4 - Beta',
        'Topic :: Internet',
        'Topic :: Database',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
    ],
)



from dbparti.backends import BasePartitionFilter
from dbparti.backends.exceptions import (
    PartitionRangeError,
    PartitionRangeSubtypeError,
    PartitionShowError
)


"""PostgreSQL backend partition filters for django admin"""
class PartitionFilter(BasePartitionFilter):
    """Common methods for all partition filters"""
    pass


class RangePartitionFilter(PartitionFilter):
    """Range partition filter implementation"""
    def __init__(self, *args, **kwargs):
        super(RangePartitionFilter, self).__init__(*args, **kwargs)
        self.partition_range = kwargs['partition_range']
        self.partition_subtype = kwargs['partition_subtype']

    def apply(self):
        """Dynamically loads needed partition filter depending on the partition subtype"""
        try:
            return getattr(self, '_get_{0}_filter'.format(self.partition_subtype))()
        except AttributeError:
            import re
            raise PartitionRangeSubtypeError(
                model=self.model,
                current_value=self.partition_subtype,
                allowed_values=[re.match('_get_(\w+)_filter', c).group(1) for c in dir(
                    self) if re.match('_get_\w+_filter', c) is not None]
            )

    def _get_date_filter(self):
        """Contains a partition filter for date partition subtype"""
        ranges = {
            'year': [
                "EXTRACT('year' FROM {0}.{1}) = EXTRACT('year' FROM {2})",
            ],
            'month': [
                "EXTRACT('year' FROM {0}.{1}) = EXTRACT('year' FROM NOW())",
                "EXTRACT('month' FROM {2}.{3}) = EXTRACT('month' FROM {4})",
            ],
            'week': [
                "EXTRACT('year' FROM {0}.{1}) = EXTRACT('year' FROM NOW())",
                "EXTRACT('week' FROM {2}.{3}) = EXTRACT('week' FROM {4})",
            ],
            'day': [
                "EXTRACT('year' FROM {0}.{1}) = EXTRACT('year' FROM NOW())",
                "EXTRACT('month' FROM {2}.{3}) = EXTRACT('month' FROM NOW())",
                "EXTRACT('day' FROM {4}.{5}) = EXTRACT('day' FROM {6})",
            ],
        }

        try:
            show_range = ranges[self.partition_range]
        except KeyError:
            raise PartitionRangeError(model=self.model, current_value=self.partition_range, allowed_values=ranges.keys())

        shows = {
            'current': 'NOW()',
            'previous': "NOW() - '1 {0}'::interval",
        }

        try:
            show = shows[self.partition_show]
        except KeyError:
            raise PartitionShowError(model=self.model, current_value=self.partition_show, allowed_values=shows.keys())

        return [item.format(self.table, self.partition_column, show.format(self.partition_range)) for item in show_range]
